from typing import Literal
import asyncio
import inspect
import traceback
import types
import logging

logger = logging.getLogger("yamt")


class ContextSkip(Exception):
    ...


class Grab:
    is_grabbed: bool = False

    async def __aenter__(self) -> None:
        if self.is_grabbed:
            raise ContextSkip()

        self.is_grabbed = True
        return None

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None
    ) -> Literal[True] | None:
        if isinstance(exc, ContextSkip):
            return True
        self.is_grabbed = False
        return None


# TODO: fix stack inspection in recursive call (rework locked_stack)
class StackLimitedLock(asyncio.Lock):
    stack_limit: int | None
    locked_stack: list[int]

    def __init__(self, stack_limit: int | None = None) -> None:
        logger.warning("unstable synchronization primitive used")
        self.stack_limit = stack_limit
        self.locked_stack = list()
        super().__init__()

    async def __aenter__(self) -> None:
        cur_frame = inspect.currentframe().f_back
        cur_hash = hash(cur_frame.f_code)

        if self.locked():
            for i, (frame, _) in enumerate(traceback.walk_stack(cur_frame)):
                if i == self.stack_limit:
                    break
                if hash(frame.f_code) in self.locked_stack:
                    self.locked_stack.append(cur_hash)
                    return None

        await super().__aenter__()
        self.locked_stack.append(cur_hash)
        return None

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None
    ) -> None:
        cur_frame = inspect.currentframe().f_back
        self.locked_stack.remove(hash(cur_frame.f_code))
        if self.locked():
            await super().__aexit__(exc_type, exc, tb)
        return None


class SemaphorePerSecond(asyncio.Semaphore):
    deffer_time: float
    _lock: asyncio.Lock

    def __init__(self, value: int = 1) -> None:
        super().__init__(value)
        self.deffer_time = 1 / value
        self._lock = asyncio.Lock()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None
    ) -> None:
        asyncio.create_task(self._deffered_exit(exc_type, exc, tb))
        return None

    async def _deffered_exit(self, *args, **kwargs):
        async with self._lock:
            await asyncio.sleep(self.deffer_time)
            await super().__aexit__(*args, **kwargs)



from typing import (
    TYPE_CHECKING,
    TypeVar,
    Generic,
    Iterator,
    SupportsIndex,
    Literal,
    Callable,
    Any,
    overload,
)
from collections.abc import Iterable, MutableSequence
from collections import deque
import itertools
import random

from .typing import SupportsRichComparison

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar("T")
MSeqT = TypeVar("MSeqT", bound=MutableSequence)


class ChainedSequence(MutableSequence, Generic[MSeqT, T]):
    seqs: tuple[MSeqT | MutableSequence[T]]

    @overload
    def __init__(self, *seqs: MSeqT | MutableSequence[T]) -> None:
        ...

    @overload
    def __init__(
        self,
        *seqs: Iterable[T],
        class_: type[MSeqT] | type[MutableSequence]
    ) -> None:
        ...

    def __init__(
        self,
        *seqs: MSeqT | MutableSequence[T] | Iterable[T],
        class_: type[MSeqT] | type[MutableSequence] | None = None
    ) -> None:
        assert seqs
        if class_ is not None:
            seqs = tuple(map(class_, seqs))
        self.seqs = seqs

    def __contains__(self, value: Any) -> bool:
        for i in self.seqs:
            if value in i:
                return True
        return False

    def __iter__(self) -> Iterator[T]:
        return iter(itertools.chain(*self.seqs))

    def __len__(self) -> int:
        return sum(map(len, self.seqs))

    def __reversed__(self) -> Iterator[T]:
        return reversed(tuple(iter(self)))

    @overload
    def __getitem__(self, index: SupportsIndex) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> MSeqT | MutableSequence[T]:
        ...

    def __getitem__(self, index: SupportsIndex | slice) -> T | MSeqT | MutableSequence[T]:
        index = self._resolve_index(index)
        if isinstance(index, int):
            index, seq = self._get_seq_by_overall_index(index)
            return seq[index]

        start, stop, step, rev = index
        chained = tuple(reversed(self) if rev else self)
        return self.seqs[0].__class__(chained[start:stop:step])

    @overload
    def __setitem__(self, index: SupportsIndex, value: T):
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[T]):
        ...

    def __setitem__(self, index: SupportsIndex | slice, value: T | Iterable[T]):
        index = self._resolve_index(index)
        if isinstance(index, SupportsIndex):
            index, seq = self._get_seq_by_overall_index(index)
            seq[index] = value

        start, stop, step, rev = index
        if rev:
            value = reversed(value)
        for i, cur_val in zip(range(start, stop + 1, step), value):
            self[i] = cur_val

    def __delitem__(self, index: SupportsIndex | slice):
        index = self._resolve_index(index)
        if isinstance(index, SupportsIndex):
            index, seq = self._get_seq_by_overall_index(index)
            del seq[index]
        else:
            start, stop, step, _ = index
            for i in range(start, stop + 1, step):
                del self[i]

    def __bool__(self) -> bool:
        return any(self.seqs)

    def __repr__(self) -> str:
        return f"<{self}>"

    def __str__(self) -> str:
        items_repr = ", ".join(map(repr, self.seqs))
        return f"ChainedSequence({items_repr})"

    def __format__(self, format_spec: str) -> str:
        return str(self).__format__(format_spec)

    def __copy__(self) -> "Self | MSeqT | MutableSequence[T]":
        return self.__class__(*(i.copy() for i in self.seqs))

    @property
    def specs(self) -> Iterator[tuple[int, int, MSeqT | MutableSequence[T]]]:
        """ Yields contained sequences and its start and end indices """
        index = 0
        for seq in self.seqs:
            if len(seq) == 0:
                yield index, index, seq
            else:
                yield index, (index := index + len(seq)) - 1, seq

    def index(
        self,
        value: Any,
        start: SupportsIndex = 0,
        stop: SupportsIndex = -1
    ) -> int:
        return tuple(self).index(value, start, stop)

    def count(self, value: Any) -> int:
        return tuple(self).count(value)

    def insert(self, index: SupportsIndex, value: T):
        index, seq = self._get_seq_by_overall_index(index)
        seq.insert(index, value)

    def append(self, value: T):
        self.seqs[-1].append(value)

    def extend(self, values: Iterable[T]):
        self.seqs[-1].extend(values)

    def pop(self, index: SupportsIndex = -1) -> T:
        index, seq = self._get_seq_by_overall_index(index)
        return seq.pop(index)

    def remove(self, value: T):
        for i in self.seqs:
            ln = len(i)
            i.remove(value)
            if ln != len(i):
                return

    def copy(self) -> "Self | MSeqT | MutableSequence[T]":
        return self.__copy__()

    @overload
    def reverse(self, *, only_containers: Literal[True]):
        ...

    @overload
    def reverse(self, *, only_contained: Literal[True]):
        ...

    def reverse(
        self,
        *,
        only_containers: bool = False,
        only_contained: bool = False
    ):
        assert not (only_containers and only_contained)
        if not only_containers:
            for i in self.seqs:
                i.reverse()
        if not only_contained:
            self.seqs = tuple(reversed(self.seqs))

    def clear(self):
        for i in self.seqs:
            i.clear()

    @overload
    def _resolve_index(self, index: SupportsIndex) -> int:
        ...

    @overload
    def _resolve_index(self, index: slice) -> tuple[int | None, int | None, int | None, bool]:
        ...

    @overload
    def _resolve_index(self, index: None) -> None:
        ...

    def _resolve_index(
        self,
        index: SupportsIndex | slice | None
    ) -> int | tuple[int | None, int | None, int | None, bool] | None:
        if index is None:
            return None
        ln = len(self)

        if isinstance(index, SupportsIndex):
            if not isinstance(index, int):
                index = index.__index__()
            if index < 0:
                return ln + index
            return index

        start, stop = map(
            self._resolve_index,
            (index.start, index.stop)
        )
        step = index.step
        r = False
        if step is not None and step < 0:
            r = True
            if start is not None:
                start = ln - start
            if stop is not None:
                stop = ln - stop
            step = abs(step) - 1
        return start, stop, step, r

    def _get_seq_by_overall_index(
        self,
        index: SupportsIndex
    ) -> tuple[int, MSeqT | MutableSequence[T]]:
        index = self._resolve_index(index)
        for start, end, seq in self.specs:
            if start == end:
                continue
            if start <= index <= end:
                break
        return index - start, seq  # true shit

    def shuffle(self):
        for i in self.seqs:
            if i:
                random.shuffle(i)


class ChainedList(ChainedSequence[list[T], T], Generic[T]):
    def sort(
        self,
        *,
        key: Callable[[T], SupportsRichComparison] | None = None,
        reverse: bool = False
    ):
        for i in self.seqs:
            i.sort(key=key, reverse=reverse)


class ChainedDeque(ChainedSequence[deque[T], T], Generic[T]):
    @property
    def maxlen(self) -> int | None:
        maxlens = tuple(i.maxlen for i in self.seqs)
        if not any(maxlens):
            return None
        return sum(i for i in maxlens if i is not None)

    def pop(self, index: SupportsIndex = -1) -> T:
        index, seq = self._get_seq_by_overall_index(index)
        val = seq[index]
        del seq[index]
        return val

    def popleft(self) -> T:
        for i in self.seqs:
            if i:
                break
        return i.popleft()

    def appendleft(self, value: T):
        self.seqs[-1].appendleft(value)

    def extendleft(self, value: Iterable[T]):
        self.seqs[-1].extendleft(value)

    def rotate(self, n: int):
        for i in self.seqs:
            i.rotate(n)



from typing import (
    Generator,
    TypeVar,
    Callable,
    Annotated,
    Iterator,
    Generic,
    Literal,
    Any,
    overload,
)
from collections.abc import Iterable
from collections import UserString
from enum import Enum
import itertools
import random

T = TypeVar("T")
NoneT = TypeVar("NoneT")
DefaultT = TypeVar("DefaultT")
ReturnT = TypeVar("ReturnT")
SingletonT = TypeVar("SingletonT", bound="SingletonMeta")


def recursive_base_attributes(cls: type) -> Iterator[tuple[str, Any]]:
    if cls is object:
        return

    yield from cls.__dict__.items()
    for i in cls.__bases__:
        yield from recursive_base_attributes(i)


def split_on_chuncks(size: int, *items: T) -> Generator[tuple[T, ...], None, None]:
    for i in range(0, len(items), size):
        yield items[i:i + size]


class WordForm(str, Enum):
    SINGLE = "si"
    PLURAL = "plu"
    MIXED = "mix"


class FormedWord(UserString):
    default_form: str
    si: str
    plu: str
    mix: str
    _data: str | None = None

    def __init__(
        self,
        default_form: WordForm | str = WordForm.SINGLE,
        *,
        si: str | None = None,
        plu: str | None = None,
        mix: str | None = None,
        **kwargs: str
    ) -> None:
        locs = locals().copy()
        del locs["self"], locs["default_form"], locs["kwargs"]

        if not (set(locs) <= set(WordForm)):
            raise NameError("word forms not matched with FormedWord init args")
        elif not (any(locs.values()) or any(kwargs.values())):
            self._data = default_form
            return

        self.__dict__.update((i.value, locs.get(i, "")) for i in WordForm)
        self.__dict__.update(kwargs)
        self.default_form = default_form

    @property
    def data(self) -> str:
        return self._data or getattr(self, self.default_form)


class Sentinel:
    __slots__ = ()

    def __repr__(self) -> str:
        return f"<SENTINEL ({id(self)})>"


sentinel = Sentinel()


@overload
def simple_chain(*items: Iterable[T] | T) -> Generator[T, None, None]:
    ...


@overload
def simple_chain(*items: Iterable[Any] | Any) -> Generator[Any, None, None]:
    ...


def simple_chain(*items: Iterable[Any] | Any) -> Generator[Any, None, None]:
    for i in items:
        if isinstance(i, Iterable):
            yield from i
        else:
            yield i


_mapper_sentinel = Sentinel()


def mapdefault(
    func: Callable[[T], ReturnT] | None,
    *iterables: Iterable[T] | NoneT,
    default: DefaultT | None = None,
    default_factory: Callable[[], DefaultT] | None = None,
    empty_check: bool = False,
    none: NoneT = None,
    check_values_before: bool = False,
    check_values_after: bool = False,
    weak_value_check: bool = False,
    as_tuple: bool = False
) -> tuple[ReturnT, ...] | Iterator[ReturnT] | DefaultT:
    """ Map iterables. Return default if all values is none.
        Can check not only iterables, but also values on none.
        Always calculates first element on call.
    """

    items = itertools.chain.from_iterable(
        i
        for i in iterables
        if (empty_check and i) or i != none
    )
    if func is not None:
        items = (
            _mapper_sentinel if (
                check_values_before
                and (
                    (weak_value_check and not i)
                    or i == none
                )
            ) else func(i)
            for i in items
        )

    result = _mapdefault_generator(
        items,
        none,
        check_values_after,
        weak_value_check
    )
    if not next(result):
        if default_factory is not None and callable(default_factory):
            default = default_factory()
        return default
    if as_tuple:
        result = tuple(result)
    return result


def _mapdefault_generator(
    items: "Iterable[ReturnT | NoneT | Sentinel]",
    none: NoneT = None,
    check_values_after: bool = False,
    weak_value_check: bool = False
) -> Annotated[
    Iterator[ReturnT | bool],
    bool, ReturnT, ReturnT, ...
]:
    first = True

    for i in items:
        if i is _mapper_sentinel:
            continue
        if check_values_after:
            if not (weak_value_check and i):
                continue
            elif i == none:
                continue

        if first:
            yield True
            first = False
        yield i

    if first:
        yield False


class SingletonMeta(type, Generic[SingletonT]):
    __instance: SingletonT | None = None

    def __call__(cls: type[SingletonT], *args, **kwargs) -> SingletonT:
        if cls.__instance is None:
            cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance

    def _wipe_singleton(self):
        del self.__instance
        self.__instance = None


class IterativeRandomizer(Generic[T]):
    data: list[T]

    def __init__(self, data: Iterable[T]) -> None:
        self.data = list(data)

    def get(self) -> T:
        if len(self.data) == 1:
            return self.data[0]
        else:
            i = random.randint(0, len(self.data) - 2)
            value = self.data.pop(i)
            self.data.append(value)
            return value


def anyvalue(iterable: Iterable[T]) -> T | Literal[False]:
    try:
        return next(itertools.dropwhile(lambda x: not x, iterable))
    except StopIteration:
        return False
